package de.abien.eis.api.playermgmt;

import de.abien.eis.api.persistent.Persistent;
import java.beans.PropertyChangeListener;
import java.beans.PropertyChangeSupport;
import java.security.SecureRandom;
import java.util.ArrayList;
import java.util.Calendar;
import java.util.Collection;
import java.util.Date;
import java.util.List;
import java.util.Locale;

import java.util.Random;
import javax.persistence.CascadeType;
import javax.persistence.Column;
import javax.persistence.Entity;
import javax.persistence.EntityManager;
import javax.persistence.EnumType;
import javax.persistence.Enumerated;
import javax.persistence.NamedQueries;
import javax.persistence.NamedQuery;
import javax.persistence.OneToMany;
import javax.persistence.OneToOne;
import javax.persistence.Query;
import javax.persistence.Temporal;
import javax.persistence.TemporalType;
import javax.xml.bind.annotation.XmlAccessType;
import javax.xml.bind.annotation.XmlAccessorType;
import javax.xml.bind.annotation.XmlElement;
import javax.xml.bind.annotation.XmlEnum;
import javax.xml.bind.annotation.XmlRootElement;
import javax.xml.bind.annotation.XmlTransient;

import javax.persistence.Table;

/** This class represents a player. It provides the following named queries:
 * <table border="1" cellspacing="0" cellpadding="3">
 *  <tr><th> Query Name <th> Definition <th> Parameters
 *  <tr><td> {@code Player.byUsernameAndPassword}
 *  <td> searches for a player with the given username and password
 *  <td> {@code username, password}
 *  <tr><td>{@code Player.byUsername}
 *  <td> searches for a player with the given username
 *  <td> {@code username}
 * </table>
 *
 * @author Frank Caputo
 */
@Entity
@Table(name = "PMM_PLAYER")
@NamedQueries({
    @NamedQuery(name = Player.findByUsernameAndPassword,
    query = "SELECT p FROM Player p WHERE p.username = :username AND p.password = :password"),
    @NamedQuery(name = Player.isUsernameAvailable,
    query = "SELECT p FROM Player p WHERE p.username = :username"),
    @NamedQuery(name = Player.findByUsername,
    query = "SELECT p FROM Player p WHERE p.username = :username"),
    @NamedQuery(name = Player.findByUsernameOrEmail,
    query = "SELECT p FROM Player p WHERE p.username = :username OR p.email = :email OR p.unconfirmedEmail = :email"),
    @NamedQuery(name = Player.findBySessionId,
    query = "SELECT p FROM Player p WHERE p.sessionId = :sessionId"),
    @NamedQuery(name = Player.isEmailNotInUse,
    query = "SELECT p FROM Player p WHERE (p.email = :email AND p.playerState <> :playerState) OR (p.unconfirmedEmail = :email AND p <> :player)"),
    @NamedQuery(name = Player.findByUsernameAndEmailConfirmValue,
    query = "SELECT p FROM Player p WHERE p.username = :username AND p.emailConfirmValue = :emailConfirmValue"),
    @NamedQuery(name = Player.findAll,
    query = "SELECT p FROM Player p"),
    @NamedQuery(name = Player.countAll,
    query = "SELECT COUNT(p.username) FROM Player p"),
    @NamedQuery(name = Player.findByAnyEMail,
    query = "SELECT p FROM Player p WHERE p.email = :email OR p.unconfirmedEmail = :email"),
    @NamedQuery(name = Player.findByExample,
    query = "SELECT p FROM Player p WHERE LOWER(p.lastname) LIKE :lastname")
})
@XmlAccessorType(XmlAccessType.FIELD)
@XmlRootElement(name = "Player", namespace = "de:edict:eis:api:playermgmt")
public class Player extends Persistent {

    // must be transient - would not work with JPA otherwise
    private transient PropertyChangeSupport propertyChangeSupport = new PropertyChangeSupport(this);
    @XmlTransient
    private static final long serialVersionUID = 1L;
    @XmlTransient
    private final static String PREFIX = "de.edict.eis.api.playermgmt.Player.";
    @XmlTransient
    public final static String isUsernameAvailable = PREFIX + "isUsernameAvailable";
    @XmlTransient
    public final static String findByUsername = PREFIX + "findByUsername";
    @XmlTransient
    public final static String findByUsernameOrEmail = PREFIX + "findByUsernameOrEmail";
    @XmlTransient
    public final static String findBySessionId = PREFIX + "findBySessionId";
    @XmlTransient
    public final static String isEmailNotInUse = PREFIX + "isEmailNotInUse";
    @XmlTransient
    public final static String findByEmailExcludeUsername = PREFIX + "findByEmailExcludeUsername";
    @XmlTransient
    public final static String findByUsernameAndPassword = PREFIX + "findByUsernameAndPassword";
    @XmlTransient
    public final static String findByUsernameAndEmailConfirmValue = PREFIX + "findByUsernameAndEmailConfirmValue";
    @XmlTransient
    public final static String findAll = PREFIX + "findAll";
    @XmlTransient
    public final static String countAll = PREFIX + "countAll";
    @XmlTransient
    public final static String findByAnyEMail = PREFIX + "findByAnyEMail";
    @XmlTransient
    public final static String findByExample = PREFIX + "findByExample";

    @Column(unique = true, nullable = false)
    @XmlElement
    private String username;

    @Column(nullable = false)
    @Enumerated(EnumType.STRING)
    @XmlElement
    private Title title;

    @Column(nullable = false)
    @XmlElement
    private String firstname;

    @Column(nullable = false)
    @XmlElement
    private String lastname;

    @Column(nullable = false)
    @Temporal(TemporalType.DATE)
    @XmlElement
    private Date birthdate;

    @Column(nullable = false)
    @XmlElement
    private String nationality;

    @XmlElement
    private String locale;

    @Column(nullable = false)
    @Enumerated(EnumType.STRING)
    @XmlElement
    private Currency currency;

    @Column(nullable = false)
    @Enumerated(EnumType.STRING)
    @XmlElement
    private PlayerState playerState;

    @XmlElement
    private String sessionId;

    @XmlElement
    private Boolean infoMail;

    @XmlElement
    private Boolean newsletter;
    // fields moved here from PlayerRegistration

    @Column(nullable = false)
    @XmlElement
    private String password;

    @Column(nullable = false)
    @XmlElement
    private String registrationQuestion;

    @Column(nullable = false)
    @XmlElement
    private String registrationAnswer;

    @XmlElement
    private String email;

    @XmlElement
    private String unconfirmedEmail;

    @XmlElement
    private String emailConfirmValue;

    @Column(nullable = false)
    @Temporal(TemporalType.TIMESTAMP)
    @XmlElement
    private Date registrationDate;

    @XmlElement
    private Integer loginAttempts;

    @Temporal(TemporalType.TIMESTAMP)
    @XmlElement
    private Date lastLoginAttempt;


    public void mergeRelatedEntities(EntityManager em) {
    }


    public static enum Title {

        MR, MRS;
    }

    public Title getTitle() {
        return title;
    }

    public Currency getCurrency() {
        return currency;
    }

    public String getSessionId() {
        return sessionId;
    }

    public void setSessionId(String sessionId) {
        this.sessionId = sessionId;
    }

    @XmlEnum
    public static enum PlayerState {

        /** user requested new password */
        TEMP_PASSWORD,
        /** user finished registration, but has not verified email yet */
        REGISTERED,
        /** active player with verified email */
        ACTIVE,
        /** max. number of login attempts reached */
        LOGIN_ATTEMPTS_EXPIRED,
        /** locked player thru admin activity */
        BLOCKED,
        /** player account closed thru cancel */
        CANCELED;
    }

    @XmlEnum
    public static enum Currency {

        EUR, GBP, USD;
    }

    public Player() {
    }

    /** Creates a new player with the given username.
     *
     * @param username the username
     */
    public Player(String username, String password, String registrationQuestion, String registrationAnswer, String eMail, Boolean infoMail, Boolean newsletter) {
        // TODO expect fields from PlayerRegistration here?
        if (username == null || password == null || eMail == null) {
            StringBuilder sb = new StringBuilder();
            sb.append("Empty values on construction are not allowed: username = ").append(username);
            sb.append(", password = [").append((password == null) ? "is null" : "not null, size " + password.length()).append("]");
            sb.append(", eMail = ").append(eMail);
            throw new IllegalArgumentException(sb.toString());
        }
        this.username = username;
        System.out.println("player neu");
        this.password = password;
        this.registrationQuestion = registrationQuestion;
        this.registrationAnswer = registrationAnswer;
//        this.email = eMail;
        this.unconfirmedEmail = eMail;
        this.registrationDate = new Date();
        this.infoMail = infoMail;
        this.newsletter = newsletter;
    }

    public Player(String username, String password, String registrationQuestion, String registrationAnswer, String eMail) {
        this(username, password, registrationQuestion, registrationAnswer, eMail, null, null);
    }

    /** Returns the firstname.
     *
     * @return the firstname.
     */
    public String getFirstname() {
        return firstname;
    }

    /** Returns the lastname.
     *
     * @return the lastname.
     */
    public String getLastname() {
        return lastname;
    }

    /** Returns the username.
     *
     * @return the username
     */
    public String getUsername() {
        return username;
    }

    /**
     * Calculate age of player.
     * This hides the birthday infos from the game.
     * @return player age in years
     */
    public int getAge() {
        //TODO Do we have to respect the players actual location for calculating the age? Are there different age definitions: Age at 01.01.?
        Calendar today = Calendar.getInstance();
        Calendar calDayOfBirth = Calendar.getInstance();
        calDayOfBirth.setTime(birthdate);
        int age = today.get(Calendar.YEAR) - calDayOfBirth.get(Calendar.YEAR);
        if (today.get(Calendar.DAY_OF_YEAR) < calDayOfBirth.get(Calendar.DAY_OF_YEAR)) {
            age--;
        }
        return age;
    }

    public Date getBirthdate() {
        return birthdate;
    }

    public String getNationality() {
        return nationality;
    }

    public Locale getLocale() {
        String[] localParts = locale.split("_");
        switch (localParts.length) {
            case 1:
                return new Locale(localParts[0]);
            case 2:
                return new Locale(localParts[0], localParts[1]);
            case 3:
                return new Locale(localParts[0], localParts[1], localParts[2]);
        }
        return null;
    }

    /**
     * Conveniance method returning value of address object
     * @return country of playerAddress
     */
    public String getCountry() {
        String country = null;
        return country;
    }

    public PlayerState getPlayerState() {
        return playerState;
    }

    public void setPlayerState(PlayerState playerState) {
        this.playerState = playerState;
    }

    /**
     * Conveniance method to check login status of player
     * @return true if player is logged in
     */
    public boolean isLoggedIn() {
        return sessionId != null;
    }

    public Date getRegistrationDate() {
        return registrationDate;
    }

    public boolean validatePassword(String checkPassword) {
        // TODO When using password encryption, place code here
        return password.equals(checkPassword);
    }

    public String setPasswordRandom() {
        char[] chars = new char[]{'a', 'b', 'c', 'd',
            'e', 'f', 'g', 'h', 'k', 'm', 'n', 'p', 'r', 'w', 'x', 'y',
            '2', '3', '4', '5', '6', '7', '8',};
        Random random = new SecureRandom();

        int car = chars.length - 1;

        String retval = "";
        for (int i = 0; i < 10; i++) {
            retval += chars[random.nextInt(car) + 1];
        }
        this.setPassword(retval);
        return retval;
    }

    public String getRegistrationQuestion() {
        return registrationQuestion;
    }

    public String getRegistrationAnswer() {
        return registrationAnswer;
    }

    public Date getLastLoginAttempt() {
        return lastLoginAttempt;
    }

    public void setLastLoginAttempt(Date lastLoginAttempt) {
        this.lastLoginAttempt = lastLoginAttempt;
    }

    public Integer getLoginAttempts() {
        return loginAttempts;
    }

    public void setLoginAttempts(Integer loginAttempts) {
        this.loginAttempts = loginAttempts;
    }

    public String getEmail() {
        return email != null ? email : unconfirmedEmail;
    }

    public String getConfirmedEmail() {
        return email;
    }

    /** returns if the given password is correct and clears loginAttempts
     * @param password the password
     * @return if the given password is correct
     */
    public boolean isPasswordCorrect(String password) {
        lastLoginAttempt = new Date();
        boolean passwordCorrect = false;
        if (validatePassword(password)) {
            loginAttempts = 0;
            passwordCorrect = true;
        } else {
            if (loginAttempts == null) {
                loginAttempts = 0;
            }
            loginAttempts++;
        }
        return passwordCorrect;
    }

    public String getUnconfirmedEmail() {
        return unconfirmedEmail;
    }

    public String generateEmailConfirmValue() {
        emailConfirmValue = String.valueOf((new Random()).nextInt(Integer.MAX_VALUE - 10000) + 10000);
        return emailConfirmValue;
    }

    public boolean confirmEmail(String confirmValue) {
        if (unconfirmedEmail != null && confirmValue != null && confirmValue.equals(this.emailConfirmValue)) {
            this.email = this.unconfirmedEmail;
            this.unconfirmedEmail = null;
            this.emailConfirmValue = null;
            return true;
        }
        return false;
    }

    public boolean isInfoMail() {
        return infoMail;
    }

    public void setInfoMail(boolean infoMail) {
        this.infoMail = infoMail;
    }

    public boolean isNewsletter() {
        return newsletter;
    }

    public void setNewsletter(boolean newsletter) {
        this.newsletter = newsletter;
    }

//    public void setId(long id) {
//        Object oldValue = this.id;
//        this.id = id;
//        propertyChangeSupport.firePropertyChange("id", oldValue, id);
//    }
    public void setUsername(String username) {
        Object oldValue = this.username;
        this.username = username;
        propertyChangeSupport.firePropertyChange("username", oldValue, username);
    }

    public void setTitle(Title title) {
        Object oldValue = this.title;
        this.title = title;
        propertyChangeSupport.firePropertyChange("title", oldValue, title);
    }

    public void setFirstname(String firstname) {
        Object oldValue = this.firstname;
        this.firstname = firstname;
        propertyChangeSupport.firePropertyChange("firstname", oldValue, firstname);
    }

    public void setLastname(String lastname) {
        Object oldValue = this.lastname;
        this.lastname = lastname;
        propertyChangeSupport.firePropertyChange("lastname", oldValue, lastname);
    }

    public void setBirthdate(Date birthdate) {
        Object oldValue = this.birthdate;
        this.birthdate = birthdate;
        propertyChangeSupport.firePropertyChange("birthdate", oldValue, birthdate);
    }

    public void setNationality(String nationality) {
        Object oldValue = this.nationality;
        this.nationality = nationality;
        propertyChangeSupport.firePropertyChange("nationality", oldValue, nationality);
    }

    public void setLocale(String locale) {
        Object oldValue = this.locale;
        this.locale = locale;
        propertyChangeSupport.firePropertyChange("locale", oldValue, locale);
    }

    public void setCurrency(Currency currency) {
        Object oldValue = this.currency;
        this.currency = currency;
        propertyChangeSupport.firePropertyChange("currency", oldValue, currency);
    }

    public void setPassword(String password) {
        Object oldValue = this.password;
        this.password = password;
        propertyChangeSupport.firePropertyChange("password", oldValue, password);
    }

    public void setRegistrationQuestion(String registrationQuestion) {
        Object oldValue = this.registrationQuestion;
        this.registrationQuestion = registrationQuestion;
        propertyChangeSupport.firePropertyChange("registrationQuestion", oldValue, registrationQuestion);
    }

    public void setRegistrationAnswer(String registrationAnswer) {
        Object oldValue = this.registrationAnswer;
        this.registrationAnswer = registrationAnswer;
        propertyChangeSupport.firePropertyChange("registrationAnswer", oldValue, registrationAnswer);
    }

    public void setEmail(String email) {
        Object oldValue = this.unconfirmedEmail;
        this.unconfirmedEmail = email;
        propertyChangeSupport.firePropertyChange("unconfirmedEmail", oldValue, email);
    }

    public void setUnconfirmedEmail(String unconfirmedEmail) {
        Object oldValue = this.unconfirmedEmail;
        this.unconfirmedEmail = unconfirmedEmail;
        propertyChangeSupport.firePropertyChange("unconfirmedEmail", oldValue, unconfirmedEmail);
    }

    public void addPropertyChangeListener(PropertyChangeListener listener) {
        propertyChangeSupport.addPropertyChangeListener(listener);
    }

    public void removePropertyChangeListener(PropertyChangeListener listener) {
        propertyChangeSupport.removePropertyChangeListener(listener);
    }

    public static Player createPlayer(EntityManager em, Player player) {
        //remove this search should not happen JPA Runtimeexceptions is ok
        Query query = em.createNamedQuery(Player.findByUsernameOrEmail);
        query.setParameter("username", player.getUsername());
        query.setParameter("email", player.getEmail());
        List result = query.getResultList();
        String emailConfirmValue = null;

        if (result.size() == 0) {
            player.setPlayerState(Player.PlayerState.REGISTERED);
            emailConfirmValue = player.generateEmailConfirmValue();
            em.persist(player);
            em.flush();
        } else {
            Player find = (Player) result.get(0);
            System.out.println("player: " + player + "Find: " + find);
            System.out.println("playerMail: " + player.getEmail());
        }
        return player;
    }

    public void setRegistrationDate(Date registrationDate) {
        this.registrationDate = registrationDate;
    }

}
