package de.abien.eis.playermgmt.service;

import de.abien.eis.api.playermgmt.Player;
import de.abien.eis.api.playermgmt.PlayermgmtService;
import java.util.ArrayList;
import java.util.GregorianCalendar;
import java.util.List;
import java.util.logging.Logger;
import javax.ejb.Local;
import javax.ejb.Stateless;
import javax.ejb.TransactionAttribute;
import javax.ejb.TransactionAttributeType;
import javax.persistence.EntityManager;
import javax.persistence.PersistenceContext;
import javax.persistence.PersistenceException;
import javax.persistence.Query;

/**
 *
 * @author fcap
 */
@Stateless
@Local(PlayermgmtService.class)
@TransactionAttribute(TransactionAttributeType.REQUIRES_NEW) //MANDATORY has to be always executed in an already existing transaction, because of Workflow
public class PlayermgmtServiceBean implements PlayermgmtService {
    //TODO allow lazyCleanupTime setup thru configuration

    private int lazyCleanupTime = -72;
    @PersistenceContext
    private EntityManager em;

    @Override
    public Player login(String username, String password, String sessionId) {
        Query query = em.createNamedQuery(Player.findByUsername);
        query.setParameter("username", username);
        List result = query.getResultList();

        if (result.size() == 1) {
            Player player = (Player) result.get(0);
            //TODO Lock zurücksetzen, falls Zeit abgelaufen
            if (player.getPlayerState() != Player.PlayerState.ACTIVE && player.getPlayerState() != Player.PlayerState.TEMP_PASSWORD && player.getPlayerState() != Player.PlayerState.LOGIN_ATTEMPTS_EXPIRED) {
                String msg = "Player in state " + player.getPlayerState() + ", but need to be " + Player.PlayerState.ACTIVE + " to login";
            } else {
                boolean passwordCorrect = player.isPasswordCorrect(password);
                if (passwordCorrect) {
                    player.setSessionId(sessionId);
                }
                Player merge = em.merge(player);
                return passwordCorrect ? merge : null;
            }
        }
        return null;
    }

    @Override
    public void logout(Player player) {
        player.setSessionId(null);
        try {
            em.merge(player);
            em.flush();
        } catch (PersistenceException pe) {
            Logger.getLogger(getClass().getName()).warning("Failed to logout player: " + pe.getLocalizedMessage());
        }
    }

    @Override
    public String forgotPasswordGetQuestion(String usernameOrEmail) {
        Query query = em.createNamedQuery(Player.findByUsernameOrEmail);
        query.setParameter("username", usernameOrEmail);
        query.setParameter("email", usernameOrEmail);
        query.getResultList();
        List result = query.getResultList();

        if (result.size() == 1) {
            Player player = (Player) result.get(0);

            return player.getRegistrationQuestion();
        }
        return null;
    }

    @Override
    public String forgotPassword(String usernameOrEmail, String registrationAnswer)  {
        Query query = em.createNamedQuery(Player.findByUsernameOrEmail);
        query.setParameter("username", usernameOrEmail);
        query.setParameter("email", usernameOrEmail);
        query.getResultList();
        List result = query.getResultList();

        if (result.size() == 1) {
            Player player = (Player) result.get(0);
            String pwd = player.setPasswordRandom();
            player.setLoginAttempts(null);
            player.setPlayerState(Player.PlayerState.TEMP_PASSWORD);
            Player merge = em.merge(player);
            return pwd;
        }
        return null;
    }

    @Override
    public Player updatePlayer(Player player) {
        // NOTE:
        // em.merge(player) leaves the player-reference detached.
        // if we want to do any managed operations, we have to do it on the
        // returned object.
        // we must be very careful, if relationships are invoked.
        return em.merge(player);
    }

    @Override
    public String createPlayer(Player player) {
        //remove this search should not happen JPA Runtimeexceptions is ok
        Query query = em.createNamedQuery(Player.findByUsernameOrEmail);
        query.setParameter("username", player.getUsername());
        query.setParameter("email", player.getEmail());
        List result = query.getResultList();
        String emailConfirmValue = null;

        if (result.size() == 0) {
            player.setPlayerState(Player.PlayerState.REGISTERED);
            emailConfirmValue = player.generateEmailConfirmValue();
            em.persist(player);
            em.flush();
        } else {
            Player find = (Player) result.get(0);
            System.out.println("player: " + player + "Find: " + find);
            System.out.println("playerMail: " + player.getEmail());
        }
        return emailConfirmValue;
    }

    @Override
    public String forgotEmailConfirmValue(String usernameOrEmail) {
        Query query = em.createNamedQuery(Player.findByUsernameOrEmail);
        query.setParameter("username", usernameOrEmail);
        query.setParameter("email", usernameOrEmail);
        query.getResultList();
        List result = query.getResultList();

        if (result.size() == 1) {
            Player player = (Player) result.get(0);
            String emailConfirmValue = player.generateEmailConfirmValue();
            em.merge(player);
            em.flush();
            return emailConfirmValue;
        }
        return null;
    }

    @Override
    public Player confirmEmail(String username, String emailConfirmValue) {
        Query query = em.createNamedQuery(Player.findByUsernameAndEmailConfirmValue);
        query.setParameter("username", username);
        query.setParameter("emailConfirmValue", emailConfirmValue);
        List result = query.getResultList();

        if (result.size() == 1) {
            Player player = (Player) result.get(0);
            if (player.confirmEmail(emailConfirmValue)) {
                player.setPlayerState(Player.PlayerState.ACTIVE);
            }

            Player merge = em.merge(player);
            return merge;
        }
        return null;
    }

    @Override
    public boolean isUsernameAvailable(String username) {
        Query query = em.createNamedQuery(Player.isUsernameAvailable);
        query.setParameter("username", username);
        return isAvailableLazyCleanup(query.getResultList());
    }

    @Override
    public boolean isEmailNotInUse(String email) {
        return isEmailNotInUse(null, email);
    }

    @Override
    public boolean isEmailNotInUse(Player player, String email) {
        return !isEmailInUse(player, email);
    }

    private boolean isEmailInUse(Player player, String email) {
        Query query = em.createNamedQuery(Player.findByAnyEMail);
        query.setParameter("email", email);
        List results = query.getResultList();
        results = cleanupUnconfirmedRegisteredPlayers(results);

        results.remove(player);
        return results.size() > 0;
    }

    private List<Player> cleanupUnconfirmedRegisteredPlayers(List<Player> players) {
        List cleanedPlayers = new ArrayList(players);
        for (Object playerObj : players) {
            Player player = (Player) playerObj;
            GregorianCalendar oldestAllowedReservation = new GregorianCalendar();
            oldestAllowedReservation.add(GregorianCalendar.HOUR, lazyCleanupTime);
            if (player.getPlayerState().equals(Player.PlayerState.REGISTERED) && player.getRegistrationDate() != null && player.getRegistrationDate().before(oldestAllowedReservation.getTime())) {
                deletePlayer(player);
                cleanedPlayers.remove(playerObj);
            }
        }
        return cleanedPlayers;
    }

    private boolean isAvailableLazyCleanup(List result) {
        for (Object playerObj : result) {
            Player player = (Player) playerObj;
            GregorianCalendar oldestAllowedReservation = new GregorianCalendar();
            oldestAllowedReservation.add(GregorianCalendar.HOUR, lazyCleanupTime);
            if (player.getPlayerState().equals(Player.PlayerState.REGISTERED) && player.getRegistrationDate() != null && player.getRegistrationDate().before(oldestAllowedReservation.getTime())) {
                deletePlayer(player);
            } else {
                return false;
            }
        }
        return true;
    }

    @Override
    public Player getPlayerForSessionId(String sessionId) {
        Query query = em.createNamedQuery(Player.findBySessionId);
        query.setParameter("sessionId", sessionId);
        List result = query.getResultList();

        if (result.size() == 1) {
            return (Player) result.get(0);
        }
        return null;
    }

    @Override
    public Player getPlayerForUsernameOrEmail(String usernameOrEmail) {
        Query query = em.createNamedQuery(Player.findByUsernameOrEmail);
        query.setParameter("username", usernameOrEmail);
        query.setParameter("email", usernameOrEmail);
        List result = query.getResultList();

        if (result.size() == 1) {
            return (Player) result.get(0);
        }
        return null;
    }

    @Override
    public void cancelPlayer(Player player, boolean forceCancelation) {
        if (player.getPlayerState().equals(Player.PlayerState.REGISTERED)) {
            // Player only registered, we can safely delete the player
            deletePlayer(player);
        } else if (playerCancelable(player) || forceCancelation) {
            player.setPlayerState(Player.PlayerState.CANCELED);
            updatePlayer(player);
        }
    }

    private boolean playerCancelable(Player player) {
        // TODO Add validations for player canceling: How do we want to return the reason?
        return true;
    }

    @Override
    public void deletePlayer(Player player) {
        //Entity must be managed to call remove
        Player merge = em.merge(player);
        em.remove(merge);
        em.flush();
    }

    @Override
    public Long countAllPlayers() {
        Query query = em.createNamedQuery(Player.countAll);
        return (Long) query.getSingleResult();
    }

}
